------------------------------------------------------------------------
----                                                                ----
---- WF1160 IP Core                                                 ----
----                                                                ----
---- Description:                                                   ----
---- This model provides an embedded Universal Serial Bus host      ----
---- controller compatible to the Philips ISP1160.                  ----
----                                                                ----
---- This entity is the FIFO buffer.                                ----
----                                                                ----
----                                                                ----
---- This storage buffer works as a FIFO from the perspective of    ----
---- the microcontroller. To access PTD structure, the access from  ----
---- the host controller (HC) perspective is randomly resulting in  ----
---- a linear adressing by the HC.                                  ----
---- This FIFO buffer is a 4096 Byte buffer organized in a 8 bit    ----
---- buffer width. The FIFO uses a standard one port RAM which is   ----
---- controlled as a ring buffer. The FIFO allows simultaneous      ----
---- acces from host and microcontroller side regardeless of read   ----
---- or write operation. A simultaneous read and write from either  ----
---- host or microcontroller is not supported. As we use a standard ----
---- RAM the FIFO is controlled in a pseudo dual port access using  ----
---- time multiplexing. This requires a fixed clock ratio of six    ----
---- clock cycles per access. In this way a concurrent read/write   ----
---- from the host side and from the microcontroller side is provi- ----
---- ded. For proper operation, the controlling signals for read    ----
---- and write are required to be strobes                           ----
----                                                                ----
---- The RAM is organized to handle both, the ATL (this is the      ----
---- acknowledged transfer list) and the ITL (this is the isochro-  ----
---- nous transfer list). A complete FIFO access cycle consists of  ----
---- host controller service (HC) and microcontroller service (MC). ----
---- As the RAM is 8 bit wide there are HI and LO byte accesses.    ----
---- The following services are possible:                           ----
---- MC and HC serviced, MC serviced, no HC, HC serviced, no MC and ----
---- no service. For more information see below.                    ----
----                                                                ----
----                                                                ----
---- Author(s):                                                     ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de     ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Copyright © 2020... Wolfgang Foerster - Inventronik GmbH.      ----
----                                                                ----
---- This source file may be used and distributed without           ----
---- restriction provided that this copyright statement is not      ----
---- removed from the file and that any derivative work contains    ----
---- the original copyright notice and the associated disclaimer.   ----
----                                                                ----
---- This source file is free software; you can redistribute it     ----
---- and/or modify it under the terms of the GNU Lesser General     ----
---- Public License as published by the Free Software Foundation;   ----
---- either version 2.1 of the License, or (at your option) any     ----
---- later version.                                                 ----
----                                                                ----
---- This source is distributed in the hope that it will be         ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied     ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR        ----
---- PURPOSE. See the GNU Lesser General Public License for more    ----
---- details.                                                       ----
----                                                                ----
---- You should have received a copy of the GNU Lesser General      ----
---- Public License along with this source; if not, download it     ----
---- from http://www.gnu.org/licenses/lgpl.html                     ----
----                                                                ----
------------------------------------------------------------------------
--
-- Revision History
--
-- Revision 2K20B  20201224 WF
--   Initial Release.
--

library work;
use work.wf1160_pkg.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

entity FIFO_BUFFER is
    port (
        CLK_48MHz           : in std_logic;
        RESET               : in std_logic;

        ATL_BUFF_LEN        : in std_logic_vector(11 downto 0);
        ITL_BUFF_LEN        : in std_logic_vector(11 downto 0);
        RD_ITL0_BUFF_LENGTH : buffer std_logic_vector(15 downto 0); -- Read back.
        RD_ITL1_BUFF_LENGTH : buffer std_logic_vector(15 downto 0); -- Read back.

        ATL_FIFO_FULL       : out std_logic;
        ITL0_FIFO_FULL      : out std_logic;
        ITL1_FIFO_FULL      : out std_logic;
        ATL_FIFO_EMPTY      : out std_logic;
        ITL0_FIFO_EMPTY     : out std_logic;
        ITL1_FIFO_EMPTY     : out std_logic;

        BUFFER_IN_MC        : in std_logic_vector(15 downto 0);
        BUFFER_IN_HC        : in std_logic_vector(15 downto 0);
        BUFFER_OUT_MC       : out std_logic_vector(15 downto 0);
        BUFFER_OUT_HC       : out std_logic_vector(15 downto 0);

        ATL_RD_MC           : in std_logic; -- By microprocessor.
        ATL_WR_MC           : in std_logic; -- By microprocessor.
        ITL_RD_MC           : in std_logic; -- By microprocessor.
        ITL_WR_MC           : in std_logic; -- By microprocessor.

        ATL_RD_HC           : in std_logic; -- By host controller.
        ATL_WR_HC           : in std_logic; -- By host controller.
        ITL_RD_HC           : in std_logic; -- By host controller.
        ITL_WR_HC           : in std_logic; -- By host controller.
        HC_ITL1             : in std_logic; -- By host controller.
        HC_ADR              : in std_logic_vector(11 downto 0) -- HCs FIFO address.
    );
end entity FIFO_BUFFER;

architecture BEHAVIOUR of FIFO_BUFFER is
type FIFO_TYPE is array(0 to 4095) of std_logic_vector(7 downto 0);
type FIFO_CYCLES is (IDLE, CHECK_HC, WAIT_MC1, WAIT_MC2, WAIT_HC1, WAIT_HC2, 
                     MC_ATL_WRITE_LO, MC_ATL_READ_LO, MC_ATL_WRITE_HI, MC_ATL_READ_HI, 
                     MC_ITL0_WRITE_LO, MC_ITL0_READ_LO, MC_ITL0_WRITE_HI, MC_ITL0_READ_HI, 
                     MC_ITL1_WRITE_LO, MC_ITL1_READ_LO, MC_ITL1_WRITE_HI, MC_ITL1_READ_HI, 
                     HC_ATL_WRITE_LO, HC_ATL_READ_LO, HC_ATL_WRITE_HI, HC_ATL_READ_HI, 
                     HC_ITL0_WRITE_LO, HC_ITL0_READ_LO, HC_ITL0_WRITE_HI, HC_ITL0_READ_HI, 
                     HC_ITL1_WRITE_LO, HC_ITL1_READ_LO, HC_ITL1_WRITE_HI, HC_ITL1_READ_HI);
signal FIFO_CYCLE       : FIFO_CYCLES;
signal NEXT_FIFO_CYCLE  : FIFO_CYCLES;
signal FIFO             : FIFO_TYPE;
signal FIFO_ADR         : std_logic_vector(11 downto 0);
signal FIFO_D_IN        : std_logic_vector(7 downto 0);
signal FIFO_D_OUT       : std_logic_vector(7 downto 0);
signal FIFO_WR          : std_logic;
signal ATL_HEAD         : std_logic_vector(11 downto 0);
signal ATL_TAIL         : std_logic_vector(11 downto 0);
signal ATL_RD_HC_I      : std_logic;
signal ATL_WR_HC_I      : std_logic;
signal ATL_RD_MC_I      : std_logic;
signal ATL_WR_MC_I      : std_logic;
signal ITL0_RD_MC       : std_logic;
signal ITL0_WR_MC       : std_logic;
signal ITL1_RD_MC       : std_logic;
signal ITL1_WR_MC       : std_logic;
signal ITL0_RD_HC       : std_logic;
signal ITL0_WR_HC       : std_logic;
signal ITL1_RD_HC       : std_logic;
signal ITL1_WR_HC       : std_logic;
signal ITL0_HEAD        : std_logic_vector(11 downto 0);
signal ITL0_TAIL        : std_logic_vector(11 downto 0);
signal ITL1_HEAD        : std_logic_vector(11 downto 0);
signal ITL1_TAIL        : std_logic_vector(11 downto 0);
begin

    REQUESTS: process
    -- These flip flops stores the requests of the MC or the HC.
    -- The flip flops are cleared when the request is processed.
    -- Be aware, that all control signals from the MC or the
    -- HC are strobes.
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if ATL_RD_MC = '1' then
            ATL_RD_MC_I <= '1';
        elsif FIFO_CYCLE = MC_ATL_READ_HI then
            ATL_RD_MC_I <= '0';
        end if;

        if ATL_WR_MC = '1' then
            ATL_WR_MC_I <= '1';
        elsif FIFO_CYCLE = MC_ATL_WRITE_HI then
            ATL_WR_MC_I <= '0';
        end if;

        if ATL_RD_HC = '1' then
            ATL_RD_HC_I <= '1';
        elsif FIFO_CYCLE = HC_ATL_READ_HI then
            ATL_RD_HC_I <= '0';
        end if;

        if ATL_WR_HC = '1' then
            ATL_WR_HC_I <= '1';
        elsif FIFO_CYCLE = HC_ATL_WRITE_HI then
            ATL_WR_HC_I <= '0';
        end if;

        if ITL_RD_MC = '1' and HC_ITL1 = '1' then
            ITL0_RD_MC <= '1';
        elsif FIFO_CYCLE = MC_ITL0_READ_HI then
            ITL0_RD_MC <= '0';
        end if;
            
        if ITL_RD_MC = '1' and HC_ITL1 = '0' then
            ITL1_RD_MC <= '1'; 
        elsif FIFO_CYCLE = MC_ITL1_READ_HI then
            ITL1_RD_MC <= '0';
        end if;

        if ITL_WR_MC = '1' and HC_ITL1 = '1' then
            ITL0_WR_MC <= '1';
        elsif FIFO_CYCLE = MC_ITL0_WRITE_HI then
            ITL0_WR_MC <= '0';
        end if;

        if ITL_WR_MC = '1' and HC_ITL1 = '0' then
            ITL1_WR_MC <= '1';
        elsif FIFO_CYCLE = MC_ITL1_WRITE_HI then
            ITL1_WR_MC <= '0';
        end if;

        if ITL_RD_HC = '1' and HC_ITL1 = '0' then
            ITL0_RD_HC <= '1';
        elsif FIFO_CYCLE = HC_ITL0_READ_HI then
            ITL0_RD_HC <= '0';
        end if;

        if ITL_RD_HC = '1' and HC_ITL1 = '1' then
            ITL1_RD_HC <= '1';
        elsif FIFO_CYCLE = HC_ITL1_READ_HI then
            ITL1_RD_HC <= '0';
        end if;

        if ITL_WR_HC = '1' and HC_ITL1 = '0' then
            ITL0_WR_HC <= '1';
        elsif FIFO_CYCLE = HC_ITL0_WRITE_HI then
            ITL0_WR_HC <= '0';
        end if;

        if ITL_WR_HC = '1' and HC_ITL1 = '1' then
            ITL1_WR_HC <= '1';
        elsif FIFO_CYCLE = HC_ITL1_WRITE_HI then
            ITL1_WR_HC <= '0';
        end if;
    end process REQUESTS;

    STATE_REG: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if RESET = '1' then
            FIFO_CYCLE <= IDLE;
        else
            FIFO_CYCLE <= NEXT_FIFO_CYCLE;
        end if;
    end process STATE_REG;

    -- The following table shows the four possibe access scenarios, coded in the state decoder.
    -- CYCLE:   1 ---------- 2 ---------- 3 ---------- 4 ---------- 5 ---------- 6                                  
    --        IDLE         MC_HI        MC_LO       CHECK_HC      HC_HI         HC_LO         -- MC and HC serviced.
    --        IDLE         MC_HI        MC_LO       CHECK_HC      WAIT_HC2      WAIT_HC1      -- MC serviced.       
    --        IDLE         WAIT_MC2     WAIT_MC1    CHECK_HC      HC_HI         HC_LO         -- HC serviced.       
    --        IDLE         WAIT_MC2     WAIT_MC1    CHECK_HC      WAIT_HC2      WAIT_HC1      -- No service.        

    STATE_DEC: process(FIFO_CYCLE, ATL_RD_MC_I, ATL_WR_MC_I, ITL0_RD_MC, ITL0_WR_MC, ITL1_RD_MC, ITL1_WR_MC, 
                       ATL_RD_HC_I, ATL_WR_HC_I, ITL0_RD_HC, ITL0_WR_HC, ITL1_RD_HC, ITL1_WR_HC)
    begin
        case FIFO_CYCLE is
            when IDLE =>
                if ATL_RD_MC_I = '1' then
                    NEXT_FIFO_CYCLE <= MC_ATL_READ_HI;
                elsif ATL_WR_MC_I = '1' then
                    NEXT_FIFO_CYCLE <= MC_ATL_WRITE_HI;
                elsif ITL0_RD_MC = '1' then
                    NEXT_FIFO_CYCLE <= MC_ITL0_READ_HI;
                elsif ITL0_WR_MC = '1' then
                    NEXT_FIFO_CYCLE <= MC_ITL0_WRITE_HI;
                elsif ITL1_RD_MC = '1' then
                    NEXT_FIFO_CYCLE <= MC_ITL1_READ_HI;
                elsif ITL1_WR_MC = '1' then
                    NEXT_FIFO_CYCLE <= MC_ITL1_WRITE_HI;
                elsif ATL_RD_HC_I = '1' or ATL_WR_HC_I = '1' then
                    NEXT_FIFO_CYCLE <= WAIT_MC2;
                elsif ITL0_RD_HC = '1' or ITL0_WR_HC = '1' then
                    NEXT_FIFO_CYCLE <= WAIT_MC2;
                elsif ITL1_RD_HC = '1' or ITL1_WR_HC = '1' then
                    NEXT_FIFO_CYCLE <= WAIT_MC2;
                else
                    NEXT_FIFO_CYCLE <= IDLE;
                end if;
            when WAIT_MC2 =>
                NEXT_FIFO_CYCLE <= WAIT_MC1;
            when WAIT_MC1 =>
                NEXT_FIFO_CYCLE <= CHECK_HC;
            when MC_ATL_READ_HI =>
                NEXT_FIFO_CYCLE <= MC_ATL_READ_LO;
            when MC_ATL_READ_LO =>
                NEXT_FIFO_CYCLE <= CHECK_HC;
            when MC_ATL_WRITE_HI =>
                NEXT_FIFO_CYCLE <= MC_ATL_WRITE_LO;
            when MC_ATL_WRITE_LO =>
                NEXT_FIFO_CYCLE <= CHECK_HC;
            when MC_ITL0_READ_HI =>
                NEXT_FIFO_CYCLE <= MC_ITL0_READ_LO;
            when MC_ITL0_READ_LO =>
                NEXT_FIFO_CYCLE <= CHECK_HC;
            when MC_ITL1_READ_HI =>
                NEXT_FIFO_CYCLE <= MC_ITL1_READ_LO;
            when MC_ITL1_READ_LO =>
                NEXT_FIFO_CYCLE <= CHECK_HC;
            when MC_ITL0_WRITE_HI =>
                NEXT_FIFO_CYCLE <= MC_ITL0_WRITE_LO;
            when MC_ITL0_WRITE_LO =>
                NEXT_FIFO_CYCLE <= CHECK_HC;
            when MC_ITL1_WRITE_HI =>
                NEXT_FIFO_CYCLE <= MC_ITL1_WRITE_LO;
            when MC_ITL1_WRITE_LO =>
                NEXT_FIFO_CYCLE <= CHECK_HC;
            when CHECK_HC =>
                if ATL_RD_HC_I = '1' then
                    NEXT_FIFO_CYCLE <= HC_ATL_READ_HI;
                elsif ATL_WR_HC_I = '1' then
                    NEXT_FIFO_CYCLE <= HC_ATL_WRITE_HI;
                elsif ITL0_RD_HC = '1' then
                    NEXT_FIFO_CYCLE <= HC_ITL0_READ_HI;
                elsif ITL0_WR_HC = '1' then
                    NEXT_FIFO_CYCLE <= HC_ITL0_WRITE_HI;
                elsif ITL1_RD_HC = '1' then
                    NEXT_FIFO_CYCLE <= HC_ITL1_READ_HI;
                elsif ITL1_WR_HC = '1' then
                    NEXT_FIFO_CYCLE <= HC_ITL1_WRITE_HI;
                else
                    NEXT_FIFO_CYCLE <= WAIT_HC2;
                end if;
            when HC_ATL_READ_HI =>
                NEXT_FIFO_CYCLE <= HC_ATL_READ_LO;
            when HC_ATL_READ_LO =>
                NEXT_FIFO_CYCLE <= IDLE;
            when HC_ATL_WRITE_HI =>
                NEXT_FIFO_CYCLE <= HC_ATL_WRITE_LO;
            when HC_ATL_WRITE_LO =>
                NEXT_FIFO_CYCLE <= IDLE;
            when HC_ITL0_READ_HI =>
                NEXT_FIFO_CYCLE <= HC_ITL0_READ_LO;
            when HC_ITL0_READ_LO =>
                NEXT_FIFO_CYCLE <= IDLE;
            when HC_ITL1_READ_HI =>
                NEXT_FIFO_CYCLE <= HC_ITL1_READ_LO;
            when HC_ITL1_READ_LO =>
                NEXT_FIFO_CYCLE <= IDLE;
            when HC_ITL0_WRITE_HI =>
                NEXT_FIFO_CYCLE <= HC_ITL0_WRITE_LO;
            when HC_ITL0_WRITE_LO =>
                NEXT_FIFO_CYCLE <= IDLE;
            when HC_ITL1_WRITE_HI =>
                NEXT_FIFO_CYCLE <= HC_ITL1_WRITE_LO;
            when HC_ITL1_WRITE_LO =>
                NEXT_FIFO_CYCLE <= IDLE;
            when WAIT_HC2 =>
                NEXT_FIFO_CYCLE <= WAIT_HC1;
            when WAIT_HC1 =>
                NEXT_FIFO_CYCLE <= IDLE;
        end case;
    end process STATE_DEC;

    -- The FIFO status signals:
    ATL_FIFO_EMPTY <= '1' when ATL_HEAD = ATL_TAIL else '0';
    ATL_FIFO_FULL <= '1' when ATL_HEAD >= ATL_TAIL and ATL_HEAD - ATL_TAIL >= ATL_BUFF_LEN - "10" else --  -- Number of bytes in the FIFO. Highest word will never be written.
                     '1' when ATL_HEAD - ATL_TAIL + ATL_BUFF_LEN >= ATL_BUFF_LEN - "10" else '0'; -- Number of bytes in the FIFO. Highest word will never be written.
    ITL0_FIFO_EMPTY <= '1' when ITL0_HEAD = ITL0_TAIL else '0';
    ITL0_FIFO_FULL <= '1' when ITL0_HEAD >= ITL0_TAIL and ITL0_HEAD - ITL0_TAIL >= ITL_BUFF_LEN - "10" else --  -- Number of bytes in the FIFO. Highest word will never be written.
                      '1' when ITL0_HEAD - ITL0_TAIL + ITL_BUFF_LEN >= ITL_BUFF_LEN - "10" else '0'; -- Number of bytes in the FIFO. Highest word will never be written.
    ITL1_FIFO_EMPTY <= '1' when ITL1_HEAD = ITL1_TAIL else '0';
    ITL1_FIFO_FULL <= '1' when ITL1_HEAD >= ITL1_TAIL and ITL1_HEAD - ITL1_TAIL >= ITL_BUFF_LEN - "10" else --  -- Number of bytes in the FIFO. Highest word will never be written.
                      '1' when ITL1_HEAD - ITL1_TAIL + ITL_BUFF_LEN >= ITL_BUFF_LEN - "10" else '0'; -- Number of bytes in the FIFO. Highest word will never be written.

    BUFFER_LENGTH: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if RESET = '1' then
            RD_ITL0_BUFF_LENGTH <= (others => '0');
        elsif FIFO_CYCLE = HC_ITL0_WRITE_HI then
            RD_ITL0_BUFF_LENGTH <= RD_ITL0_BUFF_LENGTH + '1';
        elsif FIFO_CYCLE = HC_ITL0_WRITE_LO then
            RD_ITL0_BUFF_LENGTH <= RD_ITL0_BUFF_LENGTH + '1';
        elsif FIFO_CYCLE = MC_ITL0_READ_HI then
            RD_ITL0_BUFF_LENGTH <= RD_ITL0_BUFF_LENGTH - '1';
        elsif FIFO_CYCLE = MC_ITL0_READ_LO then
            RD_ITL0_BUFF_LENGTH <= RD_ITL0_BUFF_LENGTH - '1';
        end if;

        if RESET = '1' then
            RD_ITL1_BUFF_LENGTH <= (others => '0');
        elsif FIFO_CYCLE = HC_ITL1_WRITE_HI then
            RD_ITL1_BUFF_LENGTH <= RD_ITL1_BUFF_LENGTH + '1';
        elsif FIFO_CYCLE = HC_ITL1_WRITE_LO then
            RD_ITL1_BUFF_LENGTH <= RD_ITL1_BUFF_LENGTH + '1';
        elsif FIFO_CYCLE = MC_ITL1_READ_HI then
            RD_ITL1_BUFF_LENGTH <= RD_ITL1_BUFF_LENGTH - '1';
        elsif FIFO_CYCLE = MC_ITL1_READ_LO then
            RD_ITL1_BUFF_LENGTH <= RD_ITL1_BUFF_LENGTH - '1';
        end if;
    end process BUFFER_LENGTH;

    ADR_POINTER: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if RESET = '1' then
            ATL_HEAD <= x"000";
            ITL0_HEAD <= x"000";
            ITL1_HEAD <= x"000";
            ATL_TAIL <= x"000";
            ITL0_TAIL <= x"000";
            ITL1_TAIL <= x"000";
        else
            if FIFO_CYCLE = MC_ATL_WRITE_HI and ATL_HEAD + "10" = ATL_BUFF_LEN then -- Rollover.
                ATL_HEAD <= x"000";
            elsif FIFO_CYCLE = HC_ATL_WRITE_HI and ATL_HEAD + "10" = ATL_BUFF_LEN then -- Rollover.
                ATL_HEAD <= x"000";
            elsif FIFO_CYCLE = MC_ATL_WRITE_HI then
                ATL_HEAD <= ATL_HEAD + "10"; -- Two bytes written.
            elsif FIFO_CYCLE = HC_ATL_WRITE_HI then
                ATL_HEAD <= ATL_HEAD + "10"; -- Two bytes written.
            end if;
    
            if FIFO_CYCLE = MC_ITL0_WRITE_HI and ITL0_HEAD + "10" = ITL_BUFF_LEN then -- Rollover.
                ITL0_HEAD <= x"000";
            elsif FIFO_CYCLE = HC_ITL0_WRITE_HI and ITL0_HEAD + "10" = ITL_BUFF_LEN then -- Rollover.
                ITL0_HEAD <= x"000";
            elsif FIFO_CYCLE = MC_ITL0_WRITE_HI then
                ITL0_HEAD <= ITL0_HEAD + "10"; -- Two bytes written.
            elsif FIFO_CYCLE = HC_ITL0_WRITE_HI then
                ITL0_HEAD <= ITL0_HEAD + "10"; -- Two bytes written.
            end if;
    
            if FIFO_CYCLE = MC_ITL1_WRITE_HI and ITL1_HEAD + "10" = ITL_BUFF_LEN then -- Rollover.
                ITL1_HEAD <= x"000";
            elsif FIFO_CYCLE = HC_ITL1_WRITE_HI and ITL1_HEAD + "10" = ITL_BUFF_LEN then -- Rollover.
                ITL1_HEAD <= x"000";
            elsif FIFO_CYCLE = MC_ITL1_WRITE_HI then
                ITL1_HEAD <= ITL1_HEAD + "10"; -- Two bytes written.
            elsif FIFO_CYCLE = HC_ITL1_WRITE_HI then
                ITL1_HEAD <= ITL1_HEAD + "10"; -- Two bytes written.
            end if;
    
            if FIFO_CYCLE = MC_ATL_READ_HI and ATL_TAIL + "10" = ATL_BUFF_LEN then -- Rollover.
                ATL_TAIL <= x"000";
            elsif FIFO_CYCLE = HC_ATL_READ_HI and ATL_TAIL + "10" = ATL_BUFF_LEN then -- Rollover.
                ATL_TAIL <= x"000";
            elsif FIFO_CYCLE = MC_ATL_READ_HI then
                ATL_TAIL <= ATL_TAIL + "10"; -- Two bytes written.
            elsif FIFO_CYCLE = HC_ATL_READ_HI then
                ATL_TAIL <= ATL_TAIL + "10"; -- Two bytes written.
            end if;
    
            if FIFO_CYCLE = MC_ITL0_READ_HI and ITL0_TAIL + "10" = ITL_BUFF_LEN then -- Rollover.
                ITL0_TAIL <= x"000";
            elsif FIFO_CYCLE = HC_ITL0_READ_HI and ITL0_TAIL + "10" = ITL_BUFF_LEN then -- Rollover.
                ITL0_TAIL <= x"000";
            elsif FIFO_CYCLE = MC_ITL0_READ_HI then
                ITL0_TAIL <= ITL0_TAIL + "10"; -- Two bytes written.
            elsif FIFO_CYCLE = HC_ITL0_READ_HI then
                ITL0_TAIL <= ITL0_TAIL + "10"; -- Two bytes written.
            end if;
    
            if FIFO_CYCLE = MC_ITL1_READ_HI and ITL1_TAIL + "10" = ITL_BUFF_LEN then -- Rollover.
                ITL1_TAIL <= x"000";
            elsif FIFO_CYCLE = HC_ITL1_READ_HI and ITL1_TAIL + "10" = ITL_BUFF_LEN then -- Rollover.
                ITL1_TAIL <= x"000";
            elsif FIFO_CYCLE = MC_ITL1_READ_HI then
                ITL1_TAIL <= ITL1_TAIL + "10"; -- Two bytes written.
            elsif FIFO_CYCLE = HC_ITL1_READ_HI then
                ITL1_TAIL <= ITL1_TAIL + "10"; -- Two bytes written.
            end if;
        end if;
    end process ADR_POINTER;

    OUTPUT_BUFFERS: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if FIFO_CYCLE = MC_ATL_READ_HI then
            BUFFER_OUT_MC(15 downto 8) <= FIFO_D_OUT;
        elsif FIFO_CYCLE = MC_ATL_READ_LO then
            BUFFER_OUT_MC(7 downto 0) <= FIFO_D_OUT;
        elsif FIFO_CYCLE = MC_ITL0_READ_HI or FIFO_CYCLE = MC_ITL1_READ_HI then
            BUFFER_OUT_MC(15 downto 8) <= FIFO_D_OUT;
        elsif FIFO_CYCLE = MC_ITL0_READ_LO or FIFO_CYCLE = MC_ITL1_READ_LO then
            BUFFER_OUT_MC(7 downto 0) <= FIFO_D_OUT;
        end if;
        
        if FIFO_CYCLE = HC_ATL_READ_HI then
            BUFFER_OUT_HC(15 downto 8) <= FIFO_D_OUT;
        elsif FIFO_CYCLE = HC_ATL_READ_LO then
            BUFFER_OUT_HC(7 downto 0) <= FIFO_D_OUT;
        elsif FIFO_CYCLE = HC_ITL0_READ_HI or FIFO_CYCLE = HC_ITL1_READ_HI then
            BUFFER_OUT_HC(15 downto 8) <= FIFO_D_OUT;
        elsif FIFO_CYCLE = HC_ITL0_READ_LO or FIFO_CYCLE = HC_ITL1_READ_LO then
            BUFFER_OUT_HC(7 downto 0) <= FIFO_D_OUT;
        end if;
    end process OUTPUT_BUFFERS;

    FIFO_ADR <= -- MC:
                ITL0_HEAD when FIFO_CYCLE = MC_ITL0_WRITE_LO else
                ITL0_HEAD + '1' when FIFO_CYCLE = MC_ITL0_WRITE_HI else
                ITL0_TAIL when FIFO_CYCLE = MC_ITL0_READ_LO else
                ITL0_TAIL + '1' when FIFO_CYCLE = MC_ITL0_READ_HI else
                ITL_BUFF_LEN + ITL1_HEAD when FIFO_CYCLE = MC_ITL1_WRITE_LO else
                ITL_BUFF_LEN + ITL1_HEAD + '1' when FIFO_CYCLE = MC_ITL1_WRITE_HI else
                ITL_BUFF_LEN + ITL1_TAIL when FIFO_CYCLE = MC_ITL1_READ_LO else
                ITL_BUFF_LEN + ITL1_TAIL + '1' when FIFO_CYCLE = MC_ITL1_READ_HI else
                (ITL_BUFF_LEN(10 downto 0) & '0') + ATL_HEAD when FIFO_CYCLE = MC_ATL_WRITE_LO else -- 2*ITL_BUFF_LEN + pointer.
                (ITL_BUFF_LEN(10 downto 0) & '0') + ATL_HEAD + '1' when FIFO_CYCLE = MC_ATL_WRITE_HI else -- 2*ITL_BUFF_LEN + pointer.
                (ITL_BUFF_LEN(10 downto 0) & '0') + ATL_TAIL when FIFO_CYCLE = MC_ATL_READ_LO else -- 2*ITL_BUFF_LEN + pointer.
                (ITL_BUFF_LEN(10 downto 0) & '0') + ATL_TAIL + '1' when FIFO_CYCLE = MC_ATL_READ_HI else -- 2*ITL_BUFF_LEN + pointer.
                -- HC:
                HC_ADR when FIFO_CYCLE = HC_ITL0_WRITE_LO else
                HC_ADR when FIFO_CYCLE = HC_ITL0_WRITE_HI else
                HC_ADR when FIFO_CYCLE = HC_ITL0_READ_LO else
                HC_ADR when FIFO_CYCLE = HC_ITL0_READ_HI else
                ITL_BUFF_LEN + HC_ADR when FIFO_CYCLE = HC_ITL1_WRITE_LO else
                ITL_BUFF_LEN + HC_ADR when FIFO_CYCLE = HC_ITL1_WRITE_HI else
                ITL_BUFF_LEN + HC_ADR when FIFO_CYCLE = HC_ITL1_READ_LO else
                ITL_BUFF_LEN + HC_ADR when FIFO_CYCLE = HC_ITL1_READ_HI else
                (ITL_BUFF_LEN(10 downto 0) & '0') + HC_ADR when FIFO_CYCLE = HC_ATL_WRITE_LO else
                (ITL_BUFF_LEN(10 downto 0) & '0') + HC_ADR when FIFO_CYCLE = HC_ATL_WRITE_HI else
                (ITL_BUFF_LEN(10 downto 0) & '0') + HC_ADR when FIFO_CYCLE = HC_ATL_READ_LO else
                (ITL_BUFF_LEN(10 downto 0) & '0') + HC_ADR;

    FIFO_D_IN <= BUFFER_IN_MC(15 downto 8) when FIFO_CYCLE = MC_ATL_WRITE_HI or FIFO_CYCLE = MC_ITL0_WRITE_HI or FIFO_CYCLE = MC_ITL1_WRITE_HI else 
                 BUFFER_IN_MC(7 downto 0) when FIFO_CYCLE = MC_ATL_WRITE_LO or FIFO_CYCLE = MC_ITL0_WRITE_LO or FIFO_CYCLE = MC_ITL1_WRITE_LO else
                 BUFFER_IN_HC(15 downto 8) when FIFO_CYCLE = HC_ATL_WRITE_HI or FIFO_CYCLE = HC_ITL0_WRITE_HI or FIFO_CYCLE = HC_ITL1_WRITE_HI else BUFFER_IN_HC(7 downto 0);
    
    FIFO_WR <= '1' when FIFO_CYCLE = MC_ATL_WRITE_LO or FIFO_CYCLE = MC_ATL_WRITE_HI or FIFO_CYCLE = MC_ITL0_WRITE_LO or FIFO_CYCLE = MC_ITL0_WRITE_HI or FIFO_CYCLE = MC_ITL1_WRITE_LO or FIFO_CYCLE = MC_ITL1_WRITE_HI else
               '1' when FIFO_CYCLE = HC_ATL_WRITE_LO or FIFO_CYCLE = HC_ATL_WRITE_HI or FIFO_CYCLE = HC_ITL0_WRITE_LO or FIFO_CYCLE = HC_ITL0_WRITE_HI or FIFO_CYCLE = HC_ITL1_WRITE_LO or FIFO_CYCLE = HC_ITL1_WRITE_HI else '0'; 

    FIFO_RAM: process(CLK_48MHz, FIFO)
    variable FIFO_ADR_PNTR  : integer range 0 to 4095;
    begin
        if CLK_48MHz = '1' and CLK_48MHz' event then
            FIFO_ADR_PNTR := To_Integer(unsigned(FIFO_ADR));
            if FIFO_WR = '1' then
                FIFO(FIFO_ADR_PNTR) <= FIFO_D_IN;
            end if;
        end if;
        FIFO_D_OUT <= FIFO(FIFO_ADR_PNTR);
    end process FIFO_RAM;
end architecture BEHAVIOUR;
