------------------------------------------------------------------------
----                                                                ----
---- WF1160 IP Core                                                 ----
----                                                                ----
---- Description:                                                   ----
---- This model provides an embedded Universal Serial Bus host      ----
---- controller compatible to the Philips ISP1160.                  ----
----                                                                ----
---- This entity is the two port root hub.                          ----
----                                                                ----
---- The host controller works with high speed signaling. This root ----
---- hub provides the signal conversion to lo speed when required.  ----
----                                                                ----
---- Lo speed device  Control sign   Hi speed device                ----
----                                                                ----
---- DP  DM           CTRL           DP  DM                         ----
----  0   0           SE0             0   0  -- Single ended zero.  ----
----  0   1            J              1   0                         ----
----  1   0            K              0   1                         ----
----  1   1           SE1             1   1  -- Single ended one.   ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
----                                                                ----
----                                                                ----
---- Author(s):                                                     ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de     ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Copyright © 2020... Wolfgang Foerster - Inventronik GmbH.      ----
----                                                                ----
---- This source file may be used and distributed without           ----
---- restriction provided that this copyright statement is not      ----
---- removed from the file and that any derivative work contains    ----
---- the original copyright notice and the associated disclaimer.   ----
----                                                                ----
---- This source file is free software; you can redistribute it     ----
---- and/or modify it under the terms of the GNU Lesser General     ----
---- Public License as published by the Free Software Foundation;   ----
---- either version 2.1 of the License, or (at your option) any     ----
---- later version.                                                 ----
----                                                                ----
---- This source is distributed in the hope that it will be         ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied     ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR        ----
---- PURPOSE. See the GNU Lesser General Public License for more    ----
---- details.                                                       ----
----                                                                ----
---- You should have received a copy of the GNU Lesser General      ----
---- Public License along with this source; if not, download it     ----
---- from http://www.gnu.org/licenses/lgpl.html                     ----
----                                                                ----
------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

entity ROOTHUB_2PORT is
    port(
        CLK_48MHz           : in std_logic;
        RESET               : in std_logic;

        DP1_IN              : in std_logic;
        DM1_IN              : in std_logic;
        DP2_IN              : in std_logic;
        DM2_IN              : in std_logic;
        DP1_OUT             : out std_logic;
        DM1_OUT             : out std_logic;
        DPM1_EN             : out std_logic;
        DP2_OUT             : out std_logic;
        DM2_OUT             : out std_logic;
        DPM2_EN             : out std_logic;

        -- USB data:
        SPEED_Tx            : in std_logic; -- '1' for high speed data.
        DP_IN               : in std_logic; -- Data line.
        DM_IN               : in std_logic; -- Data line.
        DPM_EN              : in std_logic; -- Output enable.
        DP_OUT              : out std_logic; -- Data line.
        DM_OUT              : out std_logic; -- Data line.

        -- Power switches:
        PSW1n               : buffer std_logic;
        PSW2n               : buffer std_logic;
        OC1n                : in std_logic;
        OC2n                : in std_logic;

        -- Port1 controls:
        CCS_1               : out std_logic; -- Current connect status.
        LSDA_1              : out std_logic; -- Low speed device attached.
        PESC_1              : out std_logic; -- Port enable status change.
        POCI_1              : out std_logic; -- Port overcurrent indicator.
        PRSC_1              : out std_logic; -- Port reset status change.
        PSSC_1              : buffer std_logic; -- Port suspend status change.
        PPS_1               : in std_logic; -- Port power switch.
        PRS_1               : in std_logic; -- Port reset status.

        -- Port2 controls:
        CCS_2               : out std_logic; -- Current connect status.
        LSDA_2              : out std_logic; -- Low speed device attached.
        PESC_2              : out std_logic; -- Port enable status change.
        POCI_2              : out std_logic; -- Port overcurrent indicator.
        PRSC_2              : out std_logic; -- Port reset status change.
        PSSC_2              : buffer std_logic; -- Port suspend status change.
        PPS_2               : in std_logic; -- Port power switch.
        PRS_2               : in std_logic; -- Port reset status.
        
        RHSC                : out std_logic; -- Root hub status change.
        UPSTREAM_P1_P2n     : out std_logic -- Indicates active port.
    );
end ROOTHUB_2PORT;

architecture BEHAVIOUR of ROOTHUB_2PORT is
type PORT_STATES is(UNPLUGGED, PORT_RESET, PORT_RESET_RECOVERY, PLUGGED);
signal PORT_I_STATE         : PORT_STATES;
signal PORT_I_NSTATE        : PORT_STATES;
signal PORT_II_STATE        : PORT_STATES;
signal PORT_II_NSTATE       : PORT_STATES;
signal PORT_I_ATTACHED      : boolean;
signal PORT_II_ATTACHED     : boolean;
signal P1_Txx               : std_logic;
signal P2_Txx               : std_logic;
signal SPEED_I              : std_logic;
signal SPEED_II             : std_logic;
signal STRB_1ms             : std_logic;
signal STRB_10us            : std_logic;
signal UPSTREAM_I           : std_logic;
signal UPSTREAM_II          : std_logic;
begin
    TIMING : process
    -- This is the logic providing a strobe signal
    -- every 1ms and the required Timings for USB
    -- operation.
    variable CNT_1ms    : integer range 0 to 65535;
    variable CNT_10us   : integer range 0 to 511;
    variable TIMER1     : integer range 0 to 63;
    variable TIMER2     : integer range 0 to 63;
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if CNT_10us < 479 then
            CNT_10us := CNT_10us + 1;
            STRB_10us <= '0';
        else
            CNT_10us := 0;
            STRB_10us <= '1';
        end if;
        --
        if CNT_1ms < 47999 then
            CNT_1ms := CNT_1ms + 1;
            STRB_1ms <= '0';
        else
            CNT_1ms := 0;
            STRB_1ms <= '1';
        end if;
        --
        if STRB_1ms = '1' then
            P1_Txx <= '0';
            P2_Txx <= '0';
            --
            if PORT_I_STATE /= PORT_RESET and PORT_I_NSTATE = PORT_RESET then
                TIMER1 := 0;
            elsif PORT_I_STATE /= PORT_RESET_RECOVERY and PORT_I_NSTATE = PORT_RESET_RECOVERY then
                TIMER1 := 0;
            elsif PORT_I_STATE = PORT_RESET and TIMER1 < 50 then
                TIMER1 := TIMER1 + 1;
            elsif PORT_I_STATE = PORT_RESET then
                P1_Txx <= '1';
            elsif PORT_I_STATE = PORT_RESET_RECOVERY and TIMER1 < 10 then
                TIMER1 := TIMER1 + 1;
            elsif PORT_I_STATE = PORT_RESET_RECOVERY then
                P1_Txx <= '1';
            end if;

            if PORT_II_STATE /= PORT_RESET and PORT_II_NSTATE = PORT_RESET then
                TIMER2 := 0;
            elsif PORT_II_STATE /= PORT_RESET_RECOVERY and PORT_II_NSTATE = PORT_RESET_RECOVERY then
                TIMER2 := 0;
            elsif PORT_II_STATE = PORT_RESET and TIMER2 < 50 then
                TIMER2 := TIMER2 + 1;
            elsif PORT_II_STATE = PORT_RESET then
                P2_Txx <= '1';
            elsif PORT_II_STATE = PORT_RESET_RECOVERY and TIMER2 < 10 then
                TIMER2 := TIMER2 + 1;
            elsif PORT_II_STATE = PORT_RESET_RECOVERY then
                P2_Txx <= '1';
            end if;
        end if;
    end process TIMING;

    ATTACHMENT: process
    -- This logic detects the attachment of a USB device.
    -- If the attachment is stable for 256ms, the devices
    -- are assumed attached. If the devices are attached
    -- and the signals indicate detachment for 256ms, the
    -- respective devices are assumed detached.
    variable DELAY_I    : integer range 0 to 255;
    variable DELAY_II   : integer range 0 to 255;
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if RESET = '1' then
            DELAY_I := 0;
            PORT_I_ATTACHED <= false;
        elsif PORT_I_STATE = UNPLUGGED and (DP1_IN xor DM1_IN) = '1' then
            if STRB_1ms = '1' and DELAY_I < 255 then
                DELAY_I := DELAY_I + 1;
                PORT_I_ATTACHED <= false;
            elsif STRB_1ms = '1' then
                PORT_I_ATTACHED <= true;
            end if;
        elsif PORT_I_STATE = PLUGGED and (DP1_IN or DM1_IN) = '1' then
            DELAY_I := 255;
            PORT_I_ATTACHED <= true;
        elsif PORT_I_STATE = PLUGGED and DP1_IN = '0' and DM1_IN = '0' then
            if STRB_10us = '1' and DELAY_I > 0 then -- 2.5us.
                DELAY_I := DELAY_I - 1;
                PORT_II_ATTACHED <= false;
            elsif STRB_10us = '1' then
                PORT_II_ATTACHED <= false;
            end if;
        end if;
        --
        if RESET = '1' then
            DELAY_II := 0;
            PORT_II_ATTACHED <= false;
        elsif PORT_II_STATE = UNPLUGGED and (DP2_IN xor DM2_IN) = '1' then
            if STRB_1ms = '1' and DELAY_II < 255 then
                DELAY_II := DELAY_II + 1;
                PORT_II_ATTACHED <= false;
            elsif STRB_1ms = '1' then
                PORT_II_ATTACHED <= true;
            end if;
        elsif PORT_II_STATE = PLUGGED and (DP2_IN or DM2_IN) = '1' then
            DELAY_II := 255;
            PORT_II_ATTACHED <= true;
        elsif PORT_II_STATE = PLUGGED and DP2_IN = '0' and DM2_IN = '0' then
            if STRB_10us = '1' and DELAY_II > 0 then -- 2.5us.
                DELAY_II := DELAY_II - 1;
                PORT_II_ATTACHED <= false;
            elsif STRB_10us = '1' then
                PORT_II_ATTACHED <= false;
            end if;
        end if;
    end process ATTACHMENT;

    SPEED_CTRL: process
    -- This signal switches the PHY to the respective
    -- operation mode.
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if PORT_I_STATE = UNPLUGGED and DP1_IN = '1' and DM1_IN = '0' and PORT_I_ATTACHED = true then
            SPEED_I <= '1'; -- Full speed.
        elsif PORT_I_STATE = UNPLUGGED then
            SPEED_I <= '0'; -- Lo speed.
        end if;
        --
        if PORT_II_STATE = UNPLUGGED and DP2_IN = '1' and DM2_IN = '0' and PORT_II_ATTACHED = true then
            SPEED_II <= '1'; -- Full speed.
        elsif PORT_I_STATE = UNPLUGGED then
            SPEED_II <= '0'; -- Lo speed.
        end if;
    end process SPEED_CTRL;

    USB_STATEREGS: process
    -- This is the USB state machines register.
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if RESET = '1' then
            PORT_I_STATE <= UNPLUGGED;
            PORT_II_STATE <= UNPLUGGED;
        else
            PORT_I_STATE <= PORT_I_NSTATE;
            PORT_II_STATE <= PORT_II_NSTATE;
        end if;
    end process USB_STATEREGS;

    PORT_I_STATEDEC: process(DM1_IN, DP1_IN, PORT_I_ATTACHED, P1_Txx, PORT_I_STATE, UPSTREAM_II, DPM_EN, PRS_1)
    -- This is the USB next state decoder for USB_I.
    begin
        case PORT_I_STATE is
            when UNPLUGGED =>
                if PORT_I_ATTACHED = true then
                    PORT_I_NSTATE <= PORT_RESET; -- Device detected.
                else
                    PORT_I_NSTATE <= UNPLUGGED;
                end if;
            when PORT_RESET =>
                if P1_Txx = '1' then -- 50ms.
                    PORT_I_NSTATE <= PORT_RESET_RECOVERY;
                else
                    PORT_I_NSTATE <= PORT_RESET;
                end if;
            when PORT_RESET_RECOVERY =>
                if P1_Txx = '1' and UPSTREAM_II = '0' and DPM_EN = '0' then -- 10ms and no activity on the bus.
                    PORT_I_NSTATE <= PLUGGED;
                else
                    PORT_I_NSTATE <= PORT_RESET_RECOVERY;
                end if;
            when PLUGGED =>
                if PORT_I_ATTACHED = false then -- Received detachment.
                    PORT_I_NSTATE <= UNPLUGGED; -- Disconnection detected.
                elsif PRS_1 = '1' then
                    PORT_I_NSTATE <= PORT_RESET;
                else
                    PORT_I_NSTATE <= PLUGGED;
                end if;
        end case;
    end process PORT_I_STATEDEC;

    PORT_II_STATEDEC: process(DM2_IN, DP2_IN, PORT_II_ATTACHED, P2_Txx, PORT_II_STATE, UPSTREAM_I, DPM_EN, PRS_2)
    -- This is the USB next state decoder for USB_II.
    begin
        case PORT_II_STATE is
            when UNPLUGGED =>
                if PORT_II_ATTACHED = true then
                    PORT_II_NSTATE <= PORT_RESET; -- Device detected.
                else
                    PORT_II_NSTATE <= UNPLUGGED;
                end if;
            when PORT_RESET =>
                if P2_Txx = '1' then -- 50ms.
                    PORT_II_NSTATE <= PORT_RESET_RECOVERY;
                else
                    PORT_II_NSTATE <= PORT_RESET;
                end if;
            when PORT_RESET_RECOVERY =>
                if P2_Txx = '1' and UPSTREAM_I = '0' and DPM_EN = '0' then -- 10ms and no activity on the bus.
                    PORT_II_NSTATE <= PLUGGED;
                else
                    PORT_II_NSTATE <= PORT_RESET_RECOVERY;
                end if;
            when PLUGGED =>
                if PORT_II_ATTACHED = false then -- Received detachment.
                    PORT_II_NSTATE <= UNPLUGGED; -- Disconnection detected.
                elsif PRS_2 = '1' then
                    PORT_II_NSTATE <= PORT_RESET;
                else
                    PORT_II_NSTATE <= PLUGGED;
                end if;
        end case;
    end process PORT_II_STATEDEC;

    -- Broadcast for downstreaming:
    -- The Upstream port always have high speed polarity.
    -- The reset signaling is SE0.
    DP1_OUT <= '0' when RESET = '1' else
               '0' when PORT_I_STATE = PORT_RESET else
               DP_IN when SPEED_I = '1' else not DP_IN;
    DM1_OUT <= '0' when RESET = '1' else
               '0' when PORT_I_STATE = PORT_RESET else
               DM_IN when SPEED_I = '1' else not DM_IN;
    DP2_OUT <= '0' when RESET = '1' else
               '0' when PORT_II_STATE = PORT_RESET else
               DP_IN when SPEED_I = '1' else not DP_IN;
    DM2_OUT <= '0' when RESET = '1' else
               '0' when PORT_II_STATE = PORT_RESET else
               DM_IN when SPEED_I = '1' else not DM_IN;

    DPM1_EN <= DPM_EN when PORT_I_ATTACHED = true and SPEED_I = '1' else -- Broadcast low speed and high speed data to high speed devices.
               DPM_EN when PORT_I_ATTACHED = true and SPEED_Tx = '0' else '0'; -- Broadcast low speed data to low speed devices.
    DPM2_EN <= DPM_EN when PORT_II_ATTACHED = true and SPEED_II = '1' else -- Broadcast low speed and high speed data to high speed devices.
               DPM_EN when PORT_II_ATTACHED = true and SPEED_Tx = '0' else '0'; -- Broadcast low speed data to low speed devices.

    -- The Upstream port always has high speed polarity.
    DP_OUT <= '0' when UPSTREAM_I = '1' and UPSTREAM_II = '1' else -- send K due to a collision.
              DP1_IN when UPSTREAM_I = '1' and SPEED_I = '1' else -- High speed.
              DP2_IN when UPSTREAM_II = '1' and SPEED_I = '1' else -- High speed.
              not DP1_IN when UPSTREAM_I = '1' and SPEED_I = '0' else -- Low speed.
              not DP2_IN when UPSTREAM_II = '1' and SPEED_I = '0' else -- Low speed.
              '0'; -- SE0

    DM_OUT <= '1' when UPSTREAM_I = '1' and UPSTREAM_II = '1' else -- send K due to a collision.
              DM1_IN when UPSTREAM_I = '1' and SPEED_I = '1' else -- High speed.
              DM2_IN when UPSTREAM_II = '1' and SPEED_I = '1' else -- High speed.
              not DM1_IN when UPSTREAM_I = '1' and SPEED_I = '0' else -- Low speed.
              not DM2_IN when UPSTREAM_II = '1' and SPEED_I = '0' else -- Low speed.
              '0'; -- SE0

    UPSTREAM_P1_P2n <= UPSTREAM_I;

    UPSTREAM_CTRL: process
    -- This logic provides information, when data is sent from a device to the host.
    -- The start of a transmission is detected when the J-state, which is the IDLE state,
    -- switches to the K state (start of a package, SOP). It ends, if the J-state after
    -- a detected SE0-state is detected. To avoid misfunction due to setup and hold issues,
    -- the SE0 lock is also controlled by UPSTREAM_x = '1' means may only occur in the
    -- end of a package.
    variable SE0_LOCK_I     : boolean;
    variable SE0_LOCK_II    : boolean;
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;
        if PORT_I_STATE = PLUGGED and SPEED_I = '0' and DP1_IN = '1' and DM1_IN = '0' then -- K state for Lo speed device.
            UPSTREAM_I <= '1';
            SE0_LOCK_I := false;
        elsif PORT_I_STATE = PLUGGED and SPEED_I = '1' and DP1_IN = '0' and DM1_IN = '1' then -- K state for Hi speed device.
            UPSTREAM_I <= '1';
            SE0_LOCK_I := false;
        elsif PORT_I_STATE = PLUGGED and UPSTREAM_I = '1' and DP1_IN = '0' and DM1_IN = '0' and SE0_LOCK_I = false then -- SE0 detected.
            SE0_LOCK_I := true;
        elsif PORT_I_STATE = PLUGGED and SPEED_I = '0' and DP1_IN = '0' and DM1_IN = '1' and SE0_LOCK_I = true then -- J state for Lo speed device.
            UPSTREAM_I <= '0';
        elsif PORT_I_STATE = PLUGGED and SPEED_I = '1' and DP1_IN = '1' and DM1_IN = '0' and SE0_LOCK_I = true then -- J state for Hi speed device.
            UPSTREAM_I <= '0';
        elsif PORT_I_STATE = UNPLUGGED then
            UPSTREAM_I <= '0';
        end if;
        --
        if PORT_II_STATE = PLUGGED and SPEED_II = '0' and DP2_IN = '1' and DM2_IN = '0' then -- K state for Lo speed device.
            UPSTREAM_II <= '1';
            SE0_LOCK_II := false;
        elsif PORT_II_STATE = PLUGGED and SPEED_II = '1' and DP2_IN = '0' and DM2_IN = '1' then -- K state for Hi speed device.
            UPSTREAM_II <= '1';
            SE0_LOCK_II := false;
        elsif PORT_II_STATE = PLUGGED and UPSTREAM_II = '1' and DP2_IN = '0' and DM2_IN = '0' and SE0_LOCK_II = false then -- SE0 detected.
            SE0_LOCK_II := true;
        elsif PORT_II_STATE = PLUGGED and SPEED_II = '0' and DP2_IN = '0' and DM2_IN = '1' and SE0_LOCK_II = true then -- J state for Lo speed device.
            UPSTREAM_II <= '0';
        elsif PORT_II_STATE = PLUGGED and SPEED_II = '1' and DP2_IN = '1' and DM2_IN = '0' and SE0_LOCK_II = true then -- J state for Hi speed device.
            UPSTREAM_II <= '0';
        elsif PORT_II_STATE = UNPLUGGED then
            UPSTREAM_II <= '0';
        end if;
    end process UPSTREAM_CTRL;

    POWER_SWITCHES: process
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;

        if RESET = '1' then
            PSW1n <= '1'; -- Off.
        elsif OC1n = '0' then
            PSW1n <= '1'; -- Overcurrent!
        else
            PSW1n <= not PPS_1;
        end if;

        if RESET = '1' then
            PSW2n <= '1'; -- Off.
        elsif OC2n = '0' then
            PSW2n <= '1'; -- Overcurrent!
        else
            PSW2n <= not PPS_2;
        end if;
    end process POWER_SWITCHES;

    P_RESUME: process
    -- This logic detects the 20ms K-state for each active port. This
    -- is the resume signaling condition.
    variable TIMER_1        : integer range 0 to 31;
    variable TIMER_2        : integer range 0 to 31;
    variable LOCK_1         : boolean;
    variable LOCK_2         : boolean;
    begin
        wait until CLK_48MHz = '1' and CLK_48MHz' event;

        PSSC_1 <= '0';
        PSSC_2 <= '0';

        if RESET = '1' or PORT_I_ATTACHED = false then
            TIMER_1 := 0;
            LOCK_1 := false;
        elsif SPEED_I = '0' and DP1_IN = '0' and DM1_IN = '1' then -- J-state no resume.
            TIMER_1 := 0;
            LOCK_1 := false;
        elsif SPEED_I = '1' and DP1_IN = '1' and DM1_IN = '0' then -- J-state no resume.
            TIMER_1 := 0;
            LOCK_1 := false;
        elsif STRB_1ms = '1' and SPEED_I = '0' and DP1_IN = '1' and DM1_IN = '0' and TIMER_1 < 20 then -- K-state.
            TIMER_1 := TIMER_1 + 1;
        elsif STRB_1ms = '1' and SPEED_I = '0' and DP1_IN = '1' and DM1_IN = '0' and LOCK_1 = false then -- K-state.
            PSSC_1 <= '1';
            LOCK_1 := true;
        elsif STRB_1ms = '1' and SPEED_I = '1' and DP1_IN = '0' and DM1_IN = '1' and TIMER_1 < 20 then -- K-state.
            TIMER_1 := TIMER_1 + 1;
        elsif STRB_1ms = '1' and SPEED_I = '1' and DP1_IN = '0' and DM1_IN = '1' and LOCK_1 = false then -- K-state.
            PSSC_1 <= '1';
            LOCK_1 := true;
        end if;

        if RESET = '1' or PORT_II_ATTACHED = false then
            TIMER_2 := 0;
            LOCK_2 := false;
        elsif SPEED_II = '0' and DP2_IN = '0' and DM2_IN = '1' then -- J-state no resume.
            TIMER_2 := 0;
            LOCK_2 := false;
        elsif SPEED_II = '1' and DP2_IN = '1' and DM2_IN = '0' then -- J-state no resume.
            TIMER_2 := 0;
            LOCK_2 := false;
        elsif STRB_1ms = '1' and SPEED_II = '0' and DP2_IN = '1' and DM2_IN = '0' and TIMER_2 < 20 then -- K-state.
            TIMER_2 := TIMER_2 + 1;
        elsif STRB_1ms = '1' and SPEED_II = '0' and DP2_IN = '1' and DM2_IN = '0' and LOCK_2 = false then -- K-state.
            PSSC_2 <= '1';
            LOCK_2 := true;
        elsif STRB_1ms = '1' and SPEED_II = '1' and DP2_IN = '0' and DM2_IN = '1' and TIMER_2 < 20 then -- K-state.
            TIMER_2 := TIMER_2 + 1;
        elsif STRB_1ms = '1' and SPEED_II = '1' and DP2_IN = '0' and DM2_IN = '1' and LOCK_2 = false then -- K-state.
            PSSC_2 <= '1';
            LOCK_2 := true;
        end if;
    end process P_RESUME;

    CCS_1 <= '1' when PORT_I_STATE = PLUGGED else '0';
    CCS_2 <= '1' when PORT_II_STATE = PLUGGED else '0';
    LSDA_1 <= '1' when SPEED_I = '0' else '0';
    LSDA_2 <= '1' when SPEED_II = '0' else '0';
    PRSC_1 <= '1' when PORT_I_STATE = PORT_RESET_RECOVERY and PORT_I_NSTATE = PLUGGED else '0';
    PRSC_2 <= '1' when PORT_II_STATE = PORT_RESET_RECOVERY and PORT_II_NSTATE = PLUGGED else '0';
    POCI_1 <= not OC1n;
    POCI_2 <= not OC2n;
    PESC_1 <= '1' when PSW1n = '0' and OC1n = '0' else '0';
    PESC_2 <= '1' when PSW2n = '0' and OC2n = '0' else '0';

    RHSC <= '1' when PORT_I_STATE /= PLUGGED and PORT_I_NSTATE = PLUGGED else
            '1' when PORT_I_STATE = PLUGGED and PORT_I_NSTATE /= PLUGGED else
            '1' when PORT_II_STATE /= PLUGGED and PORT_II_NSTATE = PLUGGED else
            '1' when PORT_II_STATE = PLUGGED and PORT_II_NSTATE /= PLUGGED else
            '1' when PSW1n = '0' and OC1n = '0' else
            '1' when PSW2n = '0' and OC2n = '0' else
            '1' when PSSC_1 = '1' else
            '1' when PSSC_2 = '1' else '0';
end architecture BEHAVIOUR;
